// Clock.java -- Matt Mahoney, mmahoney@cs.fit.edu

// This applet displays an analog clock with the current time.
// Illustrates applets, threads, event handling, and animation

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;

// The Timer thread forces an applet to repaint every second
class Timer extends Thread {
  private Applet applet;

  public Timer(Applet applet) {
    this.applet=applet;
  }

  // Repaint the applet every second forever
  public void run() {
    while (true) {
      try {
        sleep(1000);
      }
      catch (InterruptedException x) {}
      applet.repaint();
    }
  }
}

// The Clock applet display the current time
public class Clock extends Applet {
  private Timer timer;  // The thread that repaints every second
  private Image image;  // Off-screen image to remove flicker
  private Graphics g2;  // Graphics for the off-screen image

  // Initialize
  public void init() {
    setBackground(Color.white);
    timer=new Timer(this);
    timer.start();
    image=createImage(getSize().width, getSize().height);
    g2=image.getGraphics();

    // Catch applet resize events (normally only in appletviewer)
    addComponentListener(new ComponentAdapter() {

      // Update the size of the offscreen image and its Graphics object g2
      public void componentResized(ComponentEvent e) {
        g2.dispose();
        image=createImage(e.getComponent().getSize().width,
          e.getComponent().getSize().height);
        g2=image.getGraphics();
      }
    });
  }

  // To eliminate flicker, replace update().  The default would clear and
  // paint the onscreen image.
  public synchronized void update(Graphics g) {
    g2.setColor(getBackground());
    g2.fillRect(0, 0, getSize().width, getSize().height);
    g2.setColor(getForeground());
    paint(g);
  }

  // Draw the clock
  public void paint(Graphics g) {

    // Draw the clock face offscreen
    g2.setColor(Color.blue);
    g2.drawOval(1, 1, image.getWidth(this)-2, image.getHeight(this)-2);
    for (int i=0; i<360; i+=30)
      drawRadius(0.9, 1.0, i);  // ticks

    // Get the time and draw the clock hands offscreen.
    Calendar rightNow = Calendar.getInstance();;
    g2.setColor(Color.black);
    drawRadius(0, 0.7, rightNow.get(Calendar.SECOND)*6);  // Second hand
    drawRadius(0, 0.8, rightNow.get(Calendar.MINUTE)*6+rightNow.get(Calendar.SECOND)/10);  // Minutes
    for (int i=-2; i<=2; ++i)  // Hour hand (draw as 5 lines to make thicker)
      drawRadius(0, 0.5, rightNow.get(Calendar.HOUR_OF_DAY)*30+rightNow.get(Calendar.MINUTE)/2+i);

    // Replace the onscreen image
    g.drawImage(image, 0, 0, this);
  }

  // Draw a line from the center of the image starting at
  // radius r1 (0 to 1, fraction of the distance from the center
  // outward to r2 (0 to 1) at 0-360 degrees clockwise from top,
  // to image using g2.
  private void drawRadius(double r1, double r2, int degrees) {
    int x0=image.getWidth(this)/2;  // center
    int y0=image.getHeight(this)/2;
    double radians=(degrees-90)*2*Math.PI/360;  // 0 deg = 12:00
    double xr=Math.cos(radians);
    double yr=Math.sin(radians);
    int xr1=(int)(x0*xr*r1);  // line endpoints relative to center
    int yr1=(int)(y0*yr*r1);
    int xr2=(int)(x0*xr*r2);
    int yr2=(int)(y0*yr*r2);
    g2.drawLine(x0+xr1, y0+yr1, x0+xr2, y0+yr2);
  }

  // Dispose of g2
  public void destroy() {
    g2.dispose();
  }
}
